/* functions for communicating over a serial port.

    Copyright (C) 2007-2009 William Trevor King
    See COPYING for licence.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
    02111-1307, USA.

    The author may be contacted at <wking@drexel.edu> on the Internet,
    or write to Trevor King, Drexel University, Physics Dept., 3141
    Chestnut St., Philadelphia PA 19104, USA.

 * Best sources:
 *  http://tldp.org/HOWTO/Serial-Programming-HOWTO/
 *  http://www.easysw.com/~mike/serial/serial.html
 *  http://docs.linux.cz/programming/c/unix_examples/raw.html
 *  http://www.pololu.com/projects/prj0003/
 */

#include <stdlib.h>  /* malloc() */
#include <stdio.h>   /* Standard input/output definitions, fprintf() */
#include <string.h>  /* strlen, strncpy */
#include <unistd.h>  /* UNIX standards, close(), usleep() */
#include <fcntl.h>   /* File control definitions, open(), fcntl() */
#include <errno.h>   /* Error number definitions, errno, EAGAIN */
#include <termios.h> /* POSIX terminal control definitions */
#include "serial_io.h"

static int verbose = 0;

/* CHK() fails if (a != 0) */
#define CHK(a)                                             \
  do {                                                     \
    if( (rc = (a)) != 0 ) {                                \
      fprintf(stderr,"FAILURE: %s.%d. (%s) returned %d\n", \
              __FILE__,__LINE__, #a, rc);		   \
      fflush(stderr);                                      \
      return rc;                                           \
    }                                                      \
  } while (0)

/* AST() fails if (a == 0) */
#define AST(a)                                                   \
  do {                                                           \
    if( (rc = (a)) == 0 ) {                                      \
      fprintf(stderr, "ASSERT FAILED: %s.%d (%s) returned %d\n", \
              __FILE__,__LINE__, #a, rc);		         \
      fflush(stderr);                                            \
      return 1;                                                  \
    }                                                            \
  } while(0)

/* structure to hold all the data we need to preserve between calls */
struct serial_port_struct {
  char *device; /* Unix "/dev/ttyS0" = MS Windows COM1 port */
  int connected; /* if not true, connect() must be run */
  int fd; /* File descriptor for the port */
  struct termios old_options;
};

/* Allocate memory for a serial port structure
 * returns 0 on success
 */
int port_malloc(struct serial_port_struct **port)
{
  int rc;
  *port=(struct serial_port_struct *)malloc(sizeof(struct serial_port_struct));
  AST(*port != NULL);
  (*port)->device = NULL;
  (*port)->connected = 0;
  (*port)->fd = -1;
  return 0;
}

/* Free memory for a serial port structure
 * returns 0 on success
 */
void port_free(struct serial_port_struct **port)
{
  if ((*port)->connected)
    port_close(*port);
  if (verbose >= 4) printf("Free port %s\n", (*port)->device), fflush(stdout);
  free((*port)->device);
  free(*port);
  *port = NULL;
}

/* (re)connect to the port
 * Device string selects serial port, Unix "/dev/ttyS0" = MS Windows COM1 port.
 * Makes a local copy of the string, so you don't neet to keep yours...
 * Device is only read on the first connect, 
 * In other calls, you might as well just send NULL.
 * returns 0 if a successful new connection is made */
int port_connect(struct serial_port_struct *port, const char * const device)
{
  struct termios options;
  int rc, len;

  port->connected=0; /* reset in case we fail to connect */
  if (port->device == NULL) { /* save the device name */
    AST(device != NULL);
    len = strlen(device);
    port->device = (char *)malloc(sizeof(char)*(len+1));
    AST(port->device != NULL);
    strncpy(port->device, device, len);
    port->device[len] = '\0';
  }

  /* Open the port read/write.
   * O_NOCTTY so we don't get a controlling terminal.
   * See man 3 open, and bits/fcntl.h (a subheader of fcntl.h).
   * O_NDELAY for non-blocking reads per http://www.easysw.com/~mike/serial/serial.html#3
   */
  port->fd = open(port->device, O_RDWR | O_NOCTTY | O_NDELAY );
  if (port->fd == -1) {
    fprintf(stderr, "Error opening serial port '%s'\n", port->device);
    perror("open");
    return 1;
  }

  /* Clear FD_CLOEXEC so the file will not be closed if it is exec-ed.
   * I don't think we're executing the file, but I don't really know
   * what that means.  I'll leave this hear in case you do ;).
   */
  //AST(fcntl(port->fd, F_SETFL, 0) != -1); /* set fd status flags to 0 */

  CHK( tcgetattr(port->fd, &options) );   /* get default port options */
  port->old_options = options;   /* save a copy to restore on shutdown */

  /* go to 9600 baud, see termio.h */
  CHK( cfsetispeed(&options, B9600) ); /* set input speed */
  CHK( cfsetospeed(&options, B9600) ); /* set output speed */
  
  /* setup control modes, see termio.h */
  /* 8N1 (8 bit bytes, No parity, 1 stop bit) */
  options.c_cflag &= ~CSIZE;  /* Clear character size...                */
  options.c_cflag |= CS8;     /* ...and set it to 8 bits                */
  options.c_cflag &= ~PARENB; /* Disable parity bit                     */
  options.c_cflag &= ~CSTOPB; /* Only send one stop bit                 */
  options.c_cflag |= CLOCAL;  /* ignore modem status lines (no dialing) */
  options.c_cflag |= CREAD;   /* enable input.                          */

  /* Configure input processing */
  //options.c_iflag |= IGNPAR;  /* Ignore parity bit (but there are none...)
  options.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|INLCR|ICRNL|IGNCR|INPCK|ISTRIP|IXON);
  /* IGNBRK  ignore break condition (?)
   * BRKINT  signal interrupt on break
   * PARMRK  mark parity errors
   * INLRC   equate input NL to CR
   * ICRNL   equate input CR to NL
   * IGNCR   ignore CR                              ? why unset this ?
   * INPCK   activate input parity checking
   * ISTRIP  strip eighth bit off of input
   * IXON    activate flow control bits (start/stop)
   */

  /* Configure output processing */
  options.c_oflag &= ~OPOST;
  /* OPOST  activate all the other o_flag bits for output post proccessing */

  /* Setup input mode (line discipline) */
  options.c_lflag &= ~(ICANON | ECHO | ECHONL | ECHOE | ISIG | IEXTEN); /* raw */
  /* ICANON  buffer input lines
   * ECHO    transmits any recieved data back to the sender?
   * ECHONL  echo NLs
   * ECHOE   echo erase character as error correcting backspace
   * ISIG    generates a signal (?) when input is available ? enable signals?
   * IEXTEN  activate DISCARD, DSUSP, LNEXT, REPRINT, etc characters
   */

  /* Setup control characters */
  options.c_cc[VMIN] = 1;  /* Wait for at least 1 byte when reading */
  options.c_cc[VTIME] = 0; /* Wait indefinately                     */ 

  /* clean lines and activate the port */
  CHK( tcsetattr(port->fd, TCSAFLUSH, &options) );

  port->connected=1;
  return 0;
}

/* Close the serial port PORT */
void port_close(struct serial_port_struct *port)
{
  if(!port->connected) return;
  if (verbose >= 4) printf("Close port %s on fd %d\n", port->device, port->fd), fflush(stdout);
  if ( tcsetattr(port->fd, TCSANOW, &port->old_options) != 0 ) {
    perror("tcsetattr");
    fprintf(stderr, "Error restoring orignal parameters\n");
  }
  close(port->fd);
  port->fd = -1;
  port->connected = 0;
}

/* Write LEN bytes of DATA to the serial port PORT */
int port_write(struct serial_port_struct *port,
	       unsigned char *data,
	       size_t len)
{
  int i, rc;

  if (verbose >= 4) printf("Write to port %s on fd %d\n", port->device, port->fd), fflush(stdout);
  if (!port->connected)
    CHK(port_connect(port, NULL));
  if (verbose >= 1) { /* print output for debugging */
    //printf("%s:%d send %d:", __FILE__, __LINE__, len);
    printf("send %d     \t:", len);
    for (i=0; i<len; i++) {
      printf(" %02x", (int)data[i]);
    }
    printf("\n");
  }
  rc = write(port->fd,data,len);
  AST(rc == len);
  return 0;
}

/* read LEN bytes of DATA from the serial port PORT
 * DATA must be already allocated and of sufficient length.
 */
int port_read(struct serial_port_struct *port,
	      unsigned char *data,
	      size_t len)
{
  int i, rc;
  int zeros, max_zeros = 10; /* maximum number of times to allow a zero read */
  int num_read, total_read;
  int block_time = 1000; /* initially 1 ms */
  /* sum from n=0 to N of (A^n) = S(N)
   * in base A:
   *   A^0 = 0001
   *   A^1 = 0010
   *   A^2 = 0100
   *   ...
   * So in base 2, S(N) = 2^(N+1)-1
   *
   * 10 zero reads (N=9), doubling the wait each time,
   * will give a total wait time 2^10-1 = 1023 times
   * as big as the inital block time.
   */

  //if (verbose >= 4) printf("Read from port %s on fd %d\n", port->device, port->fd), fflush(stdout);
  if (!port->connected)
    CHK(port_connect(port, NULL));
  total_read = 0;
  zeros = 0;
  while (total_read < len && zeros < max_zeros) {
    num_read = read(port->fd,data+total_read,len-total_read);
    if (num_read <= 0) zeros++;
    if (num_read < 0) {
      if (errno == EAGAIN) {
	if (verbose >= 3)
	  printf("Error EAGAIN reading, wait %d ms and try again\n", block_time);
	usleep(block_time);
	block_time *= 2; /* wait twice as long next time around */
	continue;
      } else {
	perror("read");
	AST(num_read >= 0);
      }
    }
    total_read += num_read;
    if (verbose >= 1) {
      printf("read %d of %d\t:", total_read, len);
      for (i=0; i<total_read; i++)
	printf(" %02x", (int)data[i]);
      printf("\n");
      fflush(stdout);
    }
  }
  if ( total_read != len ) {
    if (total_read == 0) {
      printf("No data read from serial port %s.\n", port->device);
      printf("Check cabling, and test serial_io using the rs232 loopback plug\n");
    } else {
      printf("Only %d of %d read!!\n", total_read, len);
    }
    AST(total_read == len);
  }
  return 0;
}

int port_test(struct serial_port_struct *port)
{
  printf("%s:%d %s is a stub.\n", __FILE__, __LINE__, __FUNCTION__);
  return 0;
}
