/* functions for communicating over a serial port.

    Copyright (C) 2007-2009 William Trevor King
    See COPYING for licence.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
    02111-1307, USA.

    The author may be contacted at <wking@drexel.edu> on the Internet,
    or write to Trevor King, Drexel University, Physics Dept., 3141
    Chestnut St., Philadelphia PA 19104, USA.
*/

#include <stdlib.h> /* exit */
#include <stdio.h>  /* printf, fprintf, stderr */
#include <string.h> /* strlen */
#include <getopt.h> /* getopt */
#include "serial_io.h"

/* CHK() fails if (a != 0) */
#define CHK(a)                                                        \
  do {                                                                \
    if( (rc = (a)) != 0 ) {                                           \
      fprintf(stderr,"FAILURE: %s.%d. (%s) returned %d\n",__FILE__,__LINE__, #a, rc); \
      fflush(stderr);                                                 \
      goto Error;                                                     \
    }                                                                 \
  } while (0)

static const char options[] = "f:s:vh";
static const char *func_name = NULL;
static const char *description = 
 "Print a string in raw (binary) mode to the serial port\nand listen for an echo response.";

static int verbose = 0;
static char *device="/dev/ttyS0";
static char *test_string="Hi!";

static void help()
{
  printf("usage: %s [-%s]\n", func_name, options);
  puts(description);
  printf("Version %s\n\n", VERSION);
  puts("Options:");
  printf("-f <device>\tInput device (currently %s)\n", device);
  printf("-s <string>\tThe output string (currently '%s')\n", test_string);
  printf("-v\tIncrement verbose mode (currently %d)\n", verbose);
  puts("-h\tDisplay this help and exit");
  exit(1);
}

static int getargs(int argc, char **argv)
{
  int c;
  
  func_name = argv[0];

  while ((c=getopt(argc, argv, options)) != -1) {
    switch (c) {
    case 'f':
      device = optarg;
      break;
    case 's':
      test_string = optarg;
      break;
    case 'v':
      verbose++;
      break;
    case 'h':
    default:
      help();
      break;
    }
  }
  return 0;
}

int main(int argc, char **argv)
{
  serial_port_t *port=NULL;
  char input_string[80]={0};
  int rc=0, i;

  CHK( getargs(argc, argv) );

  CHK( port_malloc(&port) );
  if (verbose >= 2) printf("Connecting to %s\n", device), fflush(stdout); 
  CHK( port_connect(port, device) );
  if (verbose >= 2) printf("Writing '%s' (%d)\n", test_string, strlen(test_string)+1), fflush(stdout); 
  CHK( port_write(port, test_string, strlen(test_string)+1) );
  if (verbose >= 3) printf("Reading %d\n", strlen(test_string)+1), fflush(stdout); 
  CHK( port_read(port, input_string, strlen(test_string)+1) );
  input_string[strlen(test_string)+1] = '\0';
  if (verbose >= 2) printf("Read '%s' (%d)\n", input_string, strlen(test_string)+1), fflush(stdout); 
  for (i=0; i<strlen(test_string)+1; i++) {
    if (test_string[i] != input_string[i]) {
      rc = 1;
      printf("Difference on bit %d: sent %02x, read %02x\n",
	     i, test_string[i], input_string[i]);
    }
  }
  if (verbose >= 1) {
    if (rc == 0)
      printf("Success\n");
    else
      printf("Failure\n");
  }
  port_close(port);
  port_free(&port);
  return 0;
 Error:
  port_close(port);
  port_free(&port);
  return 1;
}
